﻿using EA;
using EmfAddin.Edit.UI;
using EmfAddin.IntegrityChecks;
using EmfAddin.IntegrityChecks.Core;
using EmfAddin.IntegrityChecks.UI;
using EmfAddin.Utils;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Xml;
using File = System.IO.File;

namespace EmfAddin
{
    [ComVisible(true)]
    public class EmfAddin
    {
        public const string NON_SPECIALIZED_ADD_IN = "";

        public const string NO_TECHNOLOGY = null;

        public const string EA_MENU_LOCATION_MAIN_MENU = "MainMenu";

        public const string EA_MENU_LOCATION_TREE_VIEW = "TreeView";

        public const string EA_MENU_LOCATION_DIAGRAM = "Diagram";

        public const string EA_TOP_LEVEL_MENU = "";

        public const string EMF_MAIN_MENU = "-EMF";

        public const string E_CLASS_STEREOTYPE_INTEGRITY_CHECK = "EClass integrity check";

        public const string E_PACKAGE_STEREOTYPE_INTEGRITY_CHECK = "EPackage integrity check";

        public const string SHOW_EMF_EDIT_VIEW = "Show EMF Edit View";

        public const string VALIDATE_MODEL = "Validate Model";

        public const string EDIT_EMF_DIALOG = "Edit EMF Properties...";

        public const int TAB_NOT_VISIBLE_AT_ALL = 0;

        public const string EMF_TAB_NAME = "EMF";

        public const string EMF_VALIDATION_OUTPUT_TAB = "EMF Model Validation";

        private EcoreEditControl emfControl;

        private EcoreEditControl emfControlWindow;

        private Repository contextRepository;

        private string contextGuid; 
            
        private ObjectType contextObjectType;

        private IDictionary<long, ObjectType> outputItemIdToTypeMapping;

        /// <summary>
        /// EA_Connect events enable Add-Ins to identify their type and to respond to Enterprise Architect start up.
        /// This event occurs when Enterprise Architect first loads your Add-In. Enterprise Architect itself is loading at this time so that while a Repository object is supplied, there is limited information that you can extract from it.
        /// The chief uses for EA_Connect are in initializing global Add-In data and for identifying the Add-In as an MDG Add-In.
        /// Also look at EA_Disconnect.
        /// </summary>
        /// <param name="repository">An Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        /// <returns>String identifying a specialized type of Add-In: 
        /// - "MDG" : MDG Add-Ins receive MDG Events and extra menu options.
        /// - "" : None-specialized Add-In.</returns>
        /// <remarks><see cref="http://www.sparxsystems.com/enterprise_architect_user_guide/13.0/automation/eaadd-insea_connect.html"/></remarks>
        public virtual string EA_Connect(Repository repository)
        {
            return NON_SPECIALIZED_ADD_IN;
        }

        /// <summary>
        /// The EA_Disconnect event enables the Add-In to respond to user requests to disconnect the model branch from an external project.
        /// This function is called when the Enterprise Architect closes. If you have stored references to Enterprise Architect objects (not particularly recommended anyway), you must release them here.
        /// In addition, .NET users must call memory management functions as shown below:
        /// GC.Collect();
        /// GC.WaitForPendingFinalizers();
        /// Also look at EA_Connect.
        /// </summary>
        public virtual void EA_Disconnect()
        {
            if (emfControl != null)
            {
                emfControl.Dispose();
                emfControl = null;
            }
            contextRepository = null;
            contextGuid = null;
        }

        /// <summary>
        /// EA_OnInitializeTechnologies requests that an Add-In pass an MDG Technology to Enterprise Architect for loading.
        /// This event occurs on Enterprise Architect startup. Return your technology XML to this function and Enterprise Architect loads and enables it.
        /// </summary>
        /// <example>
        /// public object EA_OnInitializeTechnologies(EA.Repository Repository){
        /// 	return My.Resources.MyTechnology;}
        /// </example>
        /// <param name="repository">An EA.Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        /// <returns>Return the MDG Technology as a single XML string.</returns>
        public virtual string EA_OnInitializeTechnologies(Repository repository)
        {
            Assembly assembly = Assembly.GetAssembly(GetType());
            FileVersionInfo versionInfo = FileVersionInfo.GetVersionInfo(assembly.Location);
            string mdgFile = Path.Combine(Path.GetDirectoryName(versionInfo.FileName), "MDG", "EMF MDG Technology.xml");
            if (File.Exists(mdgFile))
            {
                return File.ReadAllText(mdgFile);
            }
            else
            {
                return NO_TECHNOLOGY;
            }
        }

        /// <summary>
        /// The EA_GetMenuItems event enables the Add-In to provide the Enterprise Architect user interface with additional Add-In menu options in various context and main menus. When a user selects an Add-In menu option, an event is raised and passed back to the Add-In that originally defined that menu option.
        /// This event is raised just before Enterprise Architect has to show particular menu options to the user, and its use is described in the Define Menu Items topic.
        /// Also look at:
        /// - EA_MenuClick
        /// - EA_GetMenuState.
        /// </summary>
        /// <param name="repository">An Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        /// <param name="menuLocation">String representing the part of the user interface that brought up the menu. 
        /// Can be TreeView, MainMenu or Diagram.</param>
        /// <param name="menuName">The name of the parent menu for which sub-items are to be defined. In the case of the top-level menu it is an empty string.</param>
        /// <returns>One of the following types:
        /// - A string indicating the label for a single menu option.
        /// - An array of strings indicating a multiple menu options.
        /// - Empty (Visual Basic/VB.NET) or null (C#) to indicate that no menu should be displayed.
        /// In the case of the top-level menu it should be a single string or an array containing only one item, or Empty/null.</returns>
        public virtual object EA_GetMenuItems(Repository repository, string menuLocation, string menuName)
        {
            if (EA_MENU_LOCATION_MAIN_MENU.Equals(menuLocation))
            {
                switch (menuName)
                {
                    // defines the top level menu option
                    case EA_TOP_LEVEL_MENU:
                        return EMF_MAIN_MENU;
                    // defines the submenu options
                    case EMF_MAIN_MENU:
                        string[] subMenus = { E_PACKAGE_STEREOTYPE_INTEGRITY_CHECK/*, E_CLASS_STEREOTYPE_INTEGRITY_CHECK*/, SHOW_EMF_EDIT_VIEW, VALIDATE_MODEL };
                        return subMenus;
                }
            }
            else if (EA_MENU_LOCATION_TREE_VIEW.Equals(menuLocation) || EA_MENU_LOCATION_DIAGRAM.Equals(menuLocation))
            {
                object contextItem;
                ObjectType type = repository.GetContextItem(out contextItem);
                if (ObjectType.otPackage == type && (contextItem as Package).HasStereotype(EMFUtils.STEREOTYPE_E_PACKAGE))
                {
                    return EDIT_EMF_DIALOG;
                }
            }
            return null; // Do not display any menu
        }

        /// <summary>
        /// The EA_GetMenuState event enables the Add-In to set a particular menu option to either enabled or disabled. This is useful when dealing with locked packages and other situations where it is convenient to show a menu option, but not enable it for use.
        /// This event is raised just before Enterprise Architect has to show particular menu options to the user. Its use is described in the Define Menu Items topic.
        /// Also look at EA_GetMenuItems.
        /// </summary>
        /// <param name="repository">An Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        /// <param name="menuLocation">String representing the part of the user interface that brought up the menu. 
        /// Can be TreeView, MainMenu or Diagram.</param>
        /// <param name="menuName">The name of the parent menu for which sub-items must be defined. In the case of the top-level menu it is an empty string.</param>
        /// <param name="itemName">The name of the option actually clicked, for example, Create a New Invoice.</param>
        /// <param name="isEnabled">Boolean. Set to False to disable this particular menu option.</param>
        /// <param name="isChecked">Boolean. Set to True to check this particular menu option.</param>
        public virtual void EA_GetMenuState(Repository repository, string menuLocation, string menuName, string itemName, ref bool isEnabled, ref bool isChecked)
        {
            if (EA_MENU_LOCATION_MAIN_MENU.Equals(menuLocation) && EMF_MAIN_MENU.Equals(menuName) && E_CLASS_STEREOTYPE_INTEGRITY_CHECK.Equals(itemName))
            {
                isEnabled = repository.ConnectionString != null && repository.ConnectionString.Count() >= 1;
                isChecked = false;
            }
            else if (EA_MENU_LOCATION_MAIN_MENU.Equals(menuLocation) && EMF_MAIN_MENU.Equals(menuName) && E_PACKAGE_STEREOTYPE_INTEGRITY_CHECK.Equals(itemName))
            {
                isEnabled = repository.ConnectionString != null && repository.ConnectionString.Count() >= 1;
                isChecked = false;
            }
            else if (EA_MENU_LOCATION_MAIN_MENU.Equals(menuLocation) && EMF_MAIN_MENU.Equals(menuName) && VALIDATE_MODEL.Equals(itemName))
            {
                isEnabled = repository.ConnectionString != null && repository.ConnectionString.Count() >= 1;
                isChecked = false;
            }
        }

        /// <summary>
        /// EA_MenuClick events are received by an Add-In in response to user selection of a menu option.
        /// The event is raised when the user clicks on a particular menu option. When a user clicks on one of your non-parent menu options, your Add-In receives a MenuClick event, defined as follows:
        /// Sub EA_MenuClick(Repository As Repository, ByVal MenuName As String, ByVal ItemName As String)
        /// Notice that your code can directly access Enterprise Architect data and UI elements using Repository methods.
        /// Also look at EA_GetMenuItems.
        /// </summary>
        /// <param name="repository">An Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        /// <param name="menuLocation">String representing the part of the user interface that brought up the menu. 
        /// Can be TreeView, MainMenu or Diagram.</param>
        /// <param name="menuName">The name of the parent menu for which sub-items must be defined. In the case of the top-level menu it is an empty string.</param>
        /// <param name="itemName">The name of the option actually clicked, for example, Create a New Invoice.</param>
        public virtual void EA_MenuClick(Repository repository, string menuName, string itemName)
        {
            if (EMF_MAIN_MENU.Equals(menuName) && E_CLASS_STEREOTYPE_INTEGRITY_CHECK.Equals(itemName))
            {
                RunEClassStereotypeIntegrityCheck(repository);
            }
            else if (EMF_MAIN_MENU.Equals(menuName) && E_PACKAGE_STEREOTYPE_INTEGRITY_CHECK.Equals(itemName))
            {
                RunEPackageTaggedValuesIntegerityCheck(repository);
            }
            else if (EMF_MAIN_MENU.Equals(menuName) && SHOW_EMF_EDIT_VIEW.Equals(itemName))
            {
                ShowTabEMF(repository);
            }
            else if (EMF_MAIN_MENU.Equals(menuName) && VALIDATE_MODEL.Equals(itemName))
            {
                ValidateModel(repository);
            }
            else if (EDIT_EMF_DIALOG.Equals(itemName))
            {
                OpenEditEmfPropertiesDialog(repository);
            }
        }

        public virtual void RunEPackageTaggedValuesIntegerityCheck(Repository repository)
        {
            try
            {
                EPackageTaggedValuesIntegerityCheck check = new EPackageTaggedValuesIntegerityCheck(repository);
                IList<EPackageTaggedValueProblem> problems = check.Check();
                if (problems.Count >= 1)
                {
                    EPackageIssueForm dialog = new EPackageIssueForm(problems);
                    if (DialogResult.OK == dialog.ShowDialog(Win32Window.GetMainWindowHandle()))
                    {
                        check.Resolve();
                    }
                }
                else
                {
                    MessageBox.Show(Win32Window.GetMainWindowHandle(), "No problems found.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(Win32Window.GetMainWindowHandle(), e.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public virtual void RunEClassStereotypeIntegrityCheck(Repository repository)
        {
            try
            {
                EClassStereotypeIntegrityCheck check = new EClassStereotypeIntegrityCheck(repository);
                IList<EClassStereotypeProblem> problems = check.Check();
                if (problems.Count >= 1)
                {
                    EClassIssueForm dialog = new EClassIssueForm(problems);
                    if (DialogResult.OK == dialog.ShowDialog(Win32Window.GetMainWindowHandle()))
                    {
                        check.Resolve(problems);
                    }
                }
                else
                {
                    MessageBox.Show(Win32Window.GetMainWindowHandle(), "No problems found.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(Win32Window.GetMainWindowHandle(), e.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public virtual void OpenEditEmfPropertiesDialog(Repository repository)
        {
            object contextItem;
            ObjectType type = repository.GetContextItem(out contextItem);
            if (ObjectType.otPackage == type && (contextItem as Package).HasStereotype(EMFUtils.STEREOTYPE_E_PACKAGE))
            {
                Package package = contextItem as Package;
                EPackagePropertiesDialog dialog = new EPackagePropertiesDialog(package);
                if (DialogResult.OK == dialog.ShowDialog(Win32Window.GetMainWindowHandle()))
                {
                    repository.RefreshOpenDiagrams(false);
                    repository.ShowInProjectView(package);
                }
            }
            else
            {
                MessageBox.Show(Win32Window.GetMainWindowHandle(), "The selected element is not supported.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public virtual void ShowTabEMF(Repository repository)
        {
            try
            {
                if (TAB_NOT_VISIBLE_AT_ALL == repository.IsTabOpen(EMF_TAB_NAME))
                {
                    if (emfControl != null)
                    {
                        emfControl.Dispose();
                    }
                    emfControl = repository.AddTab(EMF_TAB_NAME, typeof(EcoreEditControl).FullName);
                    if (contextRepository != null)
                    {
                        emfControl.Show(contextRepository, contextGuid, contextObjectType);
                    }
                }
                else
                {
                    repository.ActivateTab(EMF_TAB_NAME);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(Win32Window.GetMainWindowHandle(), e.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public virtual void ValidateModel(Repository repository)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();

            openFileDialog.Filter = "exe files (*.exe)|*.exe";
            openFileDialog.FilterIndex = 0;
            openFileDialog.RestoreDirectory = true;
            openFileDialog.FileName = "EMFHeadlessValidator.exe";

            repository.CreateOutputTab(EMF_VALIDATION_OUTPUT_TAB);
            repository.EnsureOutputVisible(EMF_VALIDATION_OUTPUT_TAB);
            repository.ClearOutput(EMF_VALIDATION_OUTPUT_TAB);
            outputItemIdToTypeMapping = new Dictionary<long, ObjectType>();

            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    Process process = new Process();
                    process.StartInfo.FileName = openFileDialog.FileName;
                    process.StartInfo.WorkingDirectory = Path.GetDirectoryName(process.StartInfo.FileName);
                    process.StartInfo.Arguments = "\"" + repository.ConnectionString + "\"";
                    process.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;
                    process.StartInfo.UseShellExecute = false;
                    process.StartInfo.CreateNoWindow = true;
                    process.StartInfo.RedirectStandardOutput = true;
                    process.Start();
                    process.WaitForExit();// Waits here for the process to exit.
                    string xml = process.StandardOutput.ReadToEnd();
                    XmlDocument doc = new XmlDocument();
                    doc.LoadXml(xml);
                    foreach (XmlNode diagnostic in doc.SelectNodes("/diagnostics/diagnostic"))
                    {
                        string severity = diagnostic.GetAttributeValue("severity", string.Empty);
                        string message = diagnostic.GetAttributeValue("message", string.Empty);
                        string type = diagnostic.GetAttributeValue("type", string.Empty);
                        int id = diagnostic.GetAttributeValueInt("id", 0);
                        if ("OK".Equals(severity))
                        {
                            repository.WriteOutput(EMF_VALIDATION_OUTPUT_TAB, "The model is valid.", 0);
                        }
                        else
                        {
                            repository.WriteOutput(EMF_VALIDATION_OUTPUT_TAB, severity + ": " + message, id);
                            outputItemIdToTypeMapping.Add(id, toObjectType(type));
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Could not read file from disk. Original error: " + ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private ObjectType toObjectType(string type)
        {
            if ("package".Equals(type)) {
                return ObjectType.otPackage;
            }
            else {
                return default(ObjectType);
            }
        }

        /// <summary>
        /// EA_OnOutputItemDoubleClicked events informs Add-Ins that the user has used the mouse to double-click on a list entry in one of the user-defined output tabs.
        /// Usually an Add-In responds to this event in order to capture activity on an output tab they had previously created through a call to Repository.AddTab().
        /// Note that every loaded Add-In receives this event for every double-click on an output tab in Enterprise Architect - irrespective of whether the Add-In created that tab. Add-Ins should therefore check the TabName parameter supplied by this event to ensure that they are not responding to other Add-Ins' events.
        /// Also look at EA_OnOutputItemClicked.
        /// </summary>
        /// <param name="repository">An EA.Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        /// <param name="tabName">The name of the tab that the click occurred in. 
        /// Usually this would have been created through Repository.AddTab().</param>
        /// <param name="lineText">The text that had been supplied as the String parameter in the original call to Repository.WriteOutput().</param>
        /// <param name="id">The ID value specified in the original call to Repository.WriteOutput().</param>
        public virtual void EA_OnOutputItemDoubleClicked(Repository repository, string tabName, string lineText, long id)
        {
            if (EMF_VALIDATION_OUTPUT_TAB.Equals(tabName) && outputItemIdToTypeMapping != null)
            {
                ObjectType type = outputItemIdToTypeMapping[id];
                Object toShow = null;
                if (ObjectType.otElement.Equals(type))
                {
                    toShow = repository.GetElementByID((int)id);
                }
                else if (ObjectType.otPackage.Equals(type))
                {
                    toShow = repository.GetPackageByID((int)id);
                }
                else if (ObjectType.otAttribute.Equals(type))
                {
                    toShow = repository.GetAttributeByID((int)id);
                }
                else if (ObjectType.otMethod.Equals(type))
                {
                    toShow = repository.GetMethodByID((int)id);
                }
                else if (ObjectType.otConnector.Equals(type))
                {
                    Connector connector = repository.GetConnectorByID((int)id);
                    toShow = repository.GetElementByID(connector.ClientID);
                }
                else if (ObjectType.otDiagram.Equals(type))
                {
                    repository.OpenDiagram((int)id);
                    toShow = repository.GetDiagramByID((int)id);
                }
                if (toShow != null)
                {
                    repository.ShowInProjectView(toShow);
                }
            }
        }

        /// <summary>
        /// The EA_FileOpen event enables the Add-In to respond to a File Open event. When Enterprise Architect opens a new model file, this event is raised and passed to all Add-Ins implementing this method.
        /// The event occurs when the model being viewed by the Enterprise Architect user changes, for whatever reason (through user interaction or Add-In activity).
        /// Also look at EA_FileClose and EA_FileNew.
        /// </summary>
        /// <param name="repository">An EA.Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        public virtual void EA_FileOpen(Repository repository)
        {
            if (emfControlWindow == null)
            {
                emfControlWindow = repository.AddWindow(EMF_TAB_NAME, typeof(EcoreEditControl).FullName);
            }
        }

        /// <summary>
        /// The EA_FileClose event enables the Add-In to respond to a File Close event. When Enterprise Architect closes an opened Model file, this event is raised and passed to all Add-Ins implementing this method.
        /// This event occurs when the model currently opened within Enterprise Architect is about to be closed (when another model is about to be opened or when Enterprise Architect is about to shutdown).
        /// Also look at EA_FileOpen and EA_FileNew.
        /// </summary>
        /// <param name="repository">An EA.Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        public virtual void EA_FileClose(Repository repository)
        {
            repository.ClearOutput(EMF_VALIDATION_OUTPUT_TAB);
            outputItemIdToTypeMapping = null;
        }

        /// <summary>
        /// EA_OnContextItemChanged notifies Add-Ins that a different item is now in context.
        /// This event occurs after a user has selected an item anywhere in the Enterprise Architect GUI. Add-Ins that require knowledge of the current item in context can subscribe to this broadcast function. If ot = otRepository, then this function behaves the same as EA_FileOpen.
        /// Also look at EA_OnContextItemDoubleClicked and EA_OnNotifyContextItemModified.
        /// </summary>
        /// <param name="repository">An EA.Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        /// <param name="guid">Contains the GUID of the new context item. 
        /// This value corresponds to the following properties, depending on the value of the ot parameter:
        /// ot (ObjectType)	- GUID value
        /// otElement  		- Element.ElementGUID
        /// otPackage 		- Package.PackageGUID
        /// otDiagram		- Diagram.DiagramGUID
        /// otAttribute		- Attribute.AttributeGUID
        /// otMethod		- Method.MethodGUID
        /// otConnector		- Connector.ConnectorGUID
        /// otRepository	- NOT APPLICABLE, GUID is an empty string
        /// </param>
        /// <param name="ot">Specifies the type of the new context item.</param>
        public virtual void EA_OnContextItemChanged(Repository repository, string guid, ObjectType ot)
        {
            contextRepository = repository;
            contextGuid = guid;
            contextObjectType = ot;
            if (emfControl != null)
            {
                emfControl.Show(repository, guid, ot);
            }
            if (emfControlWindow != null)
            {
                emfControlWindow.Show(repository, guid, ot);
            }
        }

        /// <summary>
        /// EA_OnNotifyContextItemModified notifies Add-Ins that the current context item has been modified.
        /// This event occurs when a user has modified the context item. Add-Ins that require knowledge of when an item has been modified can subscribe to this broadcast function.
        /// Also look at EA_OnContextItemChanged and EA_OnContextItemDoubleClicked.
        /// </summary>
        /// <param name="repository">An EA.Repository object representing the currently open Enterprise Architect model.
        /// Poll its members to retrieve model data and user interface status information.</param>
        /// <param name="guid">Contains the GUID of the new context item. 
        /// This value corresponds to the following properties, depending on the value of the ot parameter:
        /// ot (ObjectType)	- GUID value
        /// otElement  		- Element.ElementGUID
        /// otPackage 		- Package.PackageGUID
        /// otDiagram		- Diagram.DiagramGUID
        /// otAttribute		- Attribute.AttributeGUID
        /// otMethod		- Method.MethodGUID
        /// otConnector		- Connector.ConnectorGUID
        /// </param>
        /// <param name="ot">Specifies the type of the new context item.</param>
        public virtual void EA_OnNotifyContextItemModified(Repository repository, string guid, ObjectType ot)
        {
            if (emfControl != null)
            {
                emfControl.Show(repository, guid, ot);
            }
            if (emfControlWindow != null)
            {
                emfControlWindow.Show(repository, guid, ot);
            }
        }
    }
}
