﻿using EA;
using EmfAddin.IntegrityChecks.Core;
using EmfAddin.Utils;
using EmfAddinTest.TestUtils;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.IO;
using File = System.IO.File;

namespace EmfAddinTest
{
    [TestClass]
    public class EClassStereotypeIntegrityCheckTest
    {
        private static string tempFile;

        private static Repository repository;

        private static Package root;

        [ClassInitialize]
        public static void CreateRepository(TestContext context)
        {
            repository = new Repository();
            repository.LoadAddins();
        }

        [ClassCleanup]
        public static void DisposeRepository()
        {
            repository.CloseAddins();
            repository.Exit();
        }

        [TestInitialize]
        public void OpenEmptyRepository()
        {
            tempFile = Path.GetTempFileName() + ".EAP";
            Assert.IsTrue(repository.CreateModel(CreateModelType.cmEAPFromBase, tempFile, 0), "Failed to create new EAP file '" + tempFile + "'.");
            Assert.IsTrue(repository.OpenFile(tempFile), "Failed to open file '" + tempFile + "'.");
            root = repository.GetPackageByID(1);
            Assert.IsNotNull(root, "Root package with ID 1 does not exist.");
        }

        [TestCleanup]
        public void CloseRepository()
        {
            repository.CloseFile();
            if (tempFile != null && File.Exists(tempFile))
            {
                File.Delete(tempFile);
            }
        }

        [TestMethod]
        public void ShouldRemoveEClassFromComponent()
        {
            // given
            Package view = repository.CreatePackage(root, "ViewPackage");
            Element eclass = repository.CreateClass(view, "MyClass");
            repository.ApplyEcoreStereotype(eclass, EMFUtils.STEREOTYPE_E_CLASS);
            Element ecomponent = repository.CreateComponent(view, "MyComponent");
            repository.ApplyEcoreStereotype(ecomponent, EMFUtils.STEREOTYPE_E_CLASS);
            EClassStereotypeIntegrityCheck check = new EClassStereotypeIntegrityCheck(repository);
            IList<EClassStereotypeProblem> problems = check.Check();
            Assert.IsNotNull(problems, "No failure list returned.");
            Assert.AreEqual(1, problems.Count, "Wrong number of problems.");
            Assert.AreEqual(ecomponent.ElementID, problems[0].ElementID, "Wrong element ID at issue 0.");
            Assert.AreEqual(ecomponent.Name, problems[0].Name, "Wrong element name at issue 0.");
            Assert.AreEqual(ecomponent.Type, problems[0].Type, "Wrong element type at issue 0.");
            // when
            check.Resolve(problems);
            // then
            repository.AssertStereotype(ecomponent);
            repository.AssertStereotype(eclass, EMFUtils.STEREOTYPE_E_CLASS);
        }

        [TestMethod]
        public void ShouldReportEClassAppliedToComponent()
        {
            // given
            Package view = repository.CreatePackage(root, "ViewPackage");
            Element eclass = repository.CreateClass(view, "MyClass");
            repository.ApplyEcoreStereotype(eclass, EMFUtils.STEREOTYPE_E_CLASS);
            Element ecomponent = repository.CreateComponent(view, "MyComponent");
            repository.ApplyEcoreStereotype(ecomponent, EMFUtils.STEREOTYPE_E_CLASS);
            // when
            IList<EClassStereotypeProblem> problems = new EClassStereotypeIntegrityCheck(repository).Check();
            // then
            Assert.IsNotNull(problems, "No failure list returned.");
            Assert.AreEqual(1, problems.Count, "Wrong number of problems.");
            Assert.AreEqual(ecomponent.ElementID, problems[0].ElementID, "Wrong element ID at issue 0.");
            Assert.AreEqual(ecomponent.Name, problems[0].Name, "Wrong element name at issue 0.");
            Assert.AreEqual(ecomponent.Type, problems[0].Type, "Wrong element type at issue 0.");
        }

        [TestMethod]
        public void ShouldReportNoproblems()
        {
            // given
            Package view = repository.CreatePackage(root, "ViewPackage");
            Element eclass = repository.CreateClass(view, "MyClass");
            repository.ApplyEcoreStereotype(eclass, EMFUtils.STEREOTYPE_E_CLASS);
            Element einterface = repository.CreateInterface(view, "MyInterface");
            repository.ApplyEcoreStereotype(einterface, EMFUtils.STEREOTYPE_E_CLASS);
            // when
            IList<EClassStereotypeProblem> problems = new EClassStereotypeIntegrityCheck(repository).Check();
            // then
            Assert.IsNotNull(problems, "No failure list returned.");
            Assert.AreEqual(0, problems.Count, "Wrong number of problems.");
        }
    }
}
