﻿using EA;
using EmfAddin.Utils;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using File = System.IO.File;

namespace EmfAddinTest.TestUtils
{
    public static class EATestUtils
    {
        public static readonly string EMPTY_EA_FILE = Path.Combine("TestModels", "Empty.EAP");

        public static string BackupProjectFileInTemp(string pathInProject)
        {
            Assert.IsFalse(pathInProject == null || pathInProject.Count() == 0, "Path to file in project not defined.");
            string tmpFolder = Path.Combine(Path.GetTempPath());
            string tempModelFilePath = Path.Combine(tmpFolder, string.Format("Copy-of-{0}", Path.GetFileName(pathInProject)));
            string sourcePath = Path.Combine(GetAssemblyFolder(), pathInProject);
            Assert.IsTrue(File.Exists(sourcePath), "File '" + sourcePath + "' does not exist.");
            try
            {
                File.Copy(sourcePath, tempModelFilePath, true);
            }
            catch (Exception e)
            {
                Assert.Fail(e.Message);
            }
            Assert.IsTrue(File.Exists(tempModelFilePath), "Backup of '" + pathInProject + "' to '" + tempModelFilePath + "' failed.");
            return tempModelFilePath;
        }

        public static string GetAssemblyFolder()
        {
            string filePath = new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath;
            return Path.GetDirectoryName(filePath);
        }

        public static Package CreatePackage(this Repository repository, Package parent, string name)
        {
            Package result = parent.Packages.AddNew(name, null);
            Assert.IsTrue(result.Update(), "Failed to save created package '" + result.Name + "': " + repository.GetLastError());
            return result;
        }

        public static Element CreateClass(this Repository repository, Package package, string name)
        {
            Element component = package.Elements.AddNew(name, "Class");
            Assert.IsTrue(component.Update(), "Failed to save created class '" + component.Name + "': " + repository.GetLastError());
            return component;
        }

        public static Element CreateInterface(this Repository repository, Package package, string name)
        {
            Element component = package.Elements.AddNew(name, "Interface");
            Assert.IsTrue(component.Update(), "Failed to save created interface '" + component.Name + "': " + repository.GetLastError());
            return component;
        }

        public static Element CreateComponent(this Repository repository, Package package, string name)
        {
            Element component = package.Elements.AddNew(name, "Component");
            Assert.IsTrue(component.Update(), "Failed to save created component '" + component.Name + "': " + repository.GetLastError());
            return component;
        }

        public static void ApplyEcoreStereotype(this Repository repository, Element element, string stereotype)
        {
            IList<string> stereotypes = EAUtils.ParseStereotypes(element.StereotypeEx);
            stereotypes.Add(EMFUtils.ECORE_PROFILE + EAUtils.STEREOTYPE_SEPARATOR + stereotype);
            element.StereotypeEx = string.Join(",", stereotypes);
            Assert.IsTrue(element.Update(), "Failed to save element '" + element.Name + "': " + repository.GetLastError());
            Assert.IsTrue(repository.SynchProfile(EMFUtils.ECORE_PROFILE, stereotype), "Profile Synchronization Failed: " + repository.GetLastError());
            element.TaggedValues.Refresh();
        }

        public static void ApplyEcoreStereotype(this Repository repository, Package package, string stereotype)
        {
            IList<string> stereotypes = EAUtils.ParseStereotypes(package.StereotypeEx);
            stereotypes.Add(EMFUtils.ECORE_PROFILE + EAUtils.STEREOTYPE_SEPARATOR + stereotype);
            package.StereotypeEx = string.Join(",", stereotypes);
            Assert.IsTrue(package.Update(), "Failed to save package '" + package.Name + "': " + repository.GetLastError());
            Assert.IsTrue(repository.SynchProfile(EMFUtils.ECORE_PROFILE, stereotype), "Profile Synchronization Failed: " + repository.GetLastError());
        }

        public static void AssertStereotype(this Repository repository, Element element, params string[] expectedStereotypes)
        {
            repository.AssertStereotype(element.ElementID, expectedStereotypes);
        }

        public static void AssertStereotype(this Repository repository, int elementID, params string[] expectedStereotypes)
        {
            Element element = repository.GetElementByID(elementID);
            Assert.IsNotNull(element, "Can't find element with ID '" + elementID + "'.");
            IList<string> actual = EAUtils.ParseStereotypes(element.StereotypeEx);
            Assert.AreEqual(expectedStereotypes.Count(), actual.Count(), "Wrong number of stereotypes of " + element.Type + " '" + element.Name + " with ID " + element.ElementID + ".");
            for (int i = 0; i < expectedStereotypes.Count(); i++)
            {
                Assert.AreEqual(expectedStereotypes[i], actual[i], "Wrong stereotypes at index " + i + " of " + element.Type + " '" + element.Name + " with ID " + element.ElementID + ".");
            }
        }
    }
}
